package com.hisschemoller.model.audio.generators 
{
	import com.hisschemoller.model.AudioDriverProxy;
	import com.hisschemoller.model.audio.IAudioProcessor;
	import com.hisschemoller.utils.Conversions;
	
	import flash.utils.ByteArray;	
	/**
	 * Plays part of an MP3 file (a slice) from _startSampleIndex to _endSampleIndex.
	 * 
	 * @author Wouter Hisschemöller
	 * (c) Apr 30, 2009
	 */
	public class PitchedSamplePlayer implements IAudioProcessor 
	{
		private const BYTE_ARRAY : ByteArray = new ByteArray();
		private var _sourceByteArray : ByteArray;
		private var _deltaSamples : int;
		private var _startSampleIndex : int;
		private var _endSampleIndex : int;
		private var _currentSampleIndexNum : Number;
		private var _relativeSampleIndex : int;
		private var _attackDuration : int = 100;
		private var _decayDuration : int = 100;
		private var _gain : Number = 1;
		private var _rate : Number;

		/**
		 * @param deltaSamples : offset into the buffer where to start writing
		 * @param sourceByteArray : the source ByteArray from which to read samples.
		 * @param startSampleIndex : the sample position where the slice starts
		 * @param endSampleIndex : the sample position where the slice ends
		 * @param transpose : The number of semitones to transpose the sample.
		 */
		public function PitchedSamplePlayer(deltaSamples : int, sourceByteArray : ByteArray, startSampleIndex : int, endSampleIndex : int, transpose : Number)
		{
			_deltaSamples = deltaSamples;
			_sourceByteArray = sourceByteArray;
			_startSampleIndex = startSampleIndex;
			_endSampleIndex = endSampleIndex;
			_currentSampleIndexNum = startSampleIndex;
			_rate = Conversions.semitoneToRate(transpose);
		}

		/**
		 * @param buffer : stereo buffer of AudioDriver.BUFFER_SIZE number of samples
		 * @return Boolean : true if the playback is done and this generator can be removed
		 */
		public function process(buffer : Vector.<Vector.<Number>>) : Boolean
		{
			var scaledBufferSize : Number = AudioDriverProxy.BUFFER_SIZE * _rate;
			var currentSampleIndexInt : int = _currentSampleIndexNum;
			var alpha : Number = _currentSampleIndexNum - currentSampleIndexInt;

			var positionTargetNum : Number = alpha;
			var positionTargetInt : int = -1;

			/** Compute the number of samples needed to process one buffer (+2 for interpolation). */
			var numSamplesRequired : int = Math.ceil(scaledBufferSize) + 2;
			
			/** If the end of the slice is reached within the current buffer, adjust the samples to read. */
			if(currentSampleIndexInt + numSamplesRequired <= _endSampleIndex)
			{
				var numSamplesAvailable : int = numSamplesRequired;
			}
			else
			{
				numSamplesAvailable = _endSampleIndex - currentSampleIndexInt;
			}
			
			/** Get the samples. */
			BYTE_ARRAY.position = 0;
			BYTE_ARRAY.writeBytes(_sourceByteArray, currentSampleIndexInt * AudioDriverProxy.BYTES_PER_SAMPLE, numSamplesAvailable * AudioDriverProxy.BYTES_PER_SAMPLE);
			BYTE_ARRAY.position = 0;
			
			var l0 : Number;
			var r0 : Number;
			var l1 : Number;
			var r1 : Number;
			
			/**  */
			var n : int = (numSamplesAvailable == numSamplesRequired) ? AudioDriverProxy.BUFFER_SIZE : (numSamplesAvailable - 2) / _rate;
			
			for(var i : int = _deltaSamples;i < n; ++i)
			{
				/** A short attack to remove clicks at the start of the slice */
				if(_relativeSampleIndex + i < _attackDuration)
				{
					_gain = (_relativeSampleIndex + i) / _attackDuration;
				}
				
				/** A short decay to remove clicks at the end of the slice. */
				if(_endSampleIndex - _relativeSampleIndex + i < _decayDuration)
				{ 
					_gain = (_endSampleIndex - _relativeSampleIndex + i) / _decayDuration;
				}
				
				/** Avoid reading equal samples if _rate < 1.0. */
				if(int(positionTargetNum) != positionTargetInt)
				{
					positionTargetInt = positionTargetNum;
					
					/** Set BYTE_ARRAY read position. */
					BYTE_ARRAY.position = positionTargetInt << 3;
	
					/** Read two stereo samples (for linear interpolation). */
					l0 = BYTE_ARRAY.readFloat();
					r0 = BYTE_ARRAY.readFloat();

					l1 = BYTE_ARRAY.readFloat();
					r1 = BYTE_ARRAY.readFloat();
				}
				
				/** Add interpolated amplitudes to stream. */
				buffer[0][i] += (l0 + alpha * (l1 - l0)) * _gain;
				buffer[1][i] += (r0 + alpha * (r1 - r0)) * _gain;
				
				/** Increase target position. */
				positionTargetNum += _rate;
				
				/** Increase fraction and clamp between 0 and 1. */
				alpha += _rate;
				while(alpha >= 1.0) --alpha;
			}
			
			/** If the slice is done, fill the rest of the buffer with silence. */
			if(i < AudioDriverProxy.BUFFER_SIZE)
			{
				while(i < AudioDriverProxy.BUFFER_SIZE)
				{
					buffer[0][i] += 0.0;
					buffer[1][i] += 0.0;
					
					++i;
				}
			}
			
			_currentSampleIndexNum += scaledBufferSize - _deltaSamples;
			_relativeSampleIndex = _currentSampleIndexNum - _startSampleIndex;
			_deltaSamples = 0;
			
			if(_currentSampleIndexNum >= _endSampleIndex) return true;
			
			return false;
		}
	}
}
